# Copyright (c) 2011 Joshua D. Bartlett
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
'''
formattools: A collection of utilities which make it easier to extend the
Python string formatting.

For more details see
http://sqizit.bartletts.id.au/2011/02/07/python-context-sensitive-formatting/

We can use the TextBased class to define our own conversion and template types.

>>> import cgi
>>> class Cgi(TextBased):
...    @classmethod
...    def _convert_from_(cls, other):
...        if isinstance(other, PlainText):
...            other = raw(other)
...        if isinstance(other, (str, unicode)):
...            return Cgi(cgi.escape(other))
...        raise NotImplementedError
...    @staticmethod
...    def valid_raw_text(text):
...        return True
...

To use types, we need to register them with a formatter.

>>> f = Formatter()
>>> f.registertype(Cgi, 'cgi')
<class 'formattools.Cgi'>
>>> print f.format('In cgi, {0} is formatted as: {0:/cgi}', '3 < 5')
In cgi, 3 < 5 is formatted as: 3 &lt; 5
>>>

We can also use our newly-defined type to create templates in which fields have
the cgi field type implicitly.

>>> template = Cgi('<b>{0}</b>')
>>> result = f.format(template, '3 < 5')
>>> result
Cgi('<b>3 &lt; 5</b>')
>>> raw(result)
'<b>3 &lt; 5</b>'
>>>

If we do not specify a formatter, the default formatter is used.

>>> template.format('3 < 5')
Cgi('<b>3 &lt; 5</b>')

'''

import inspect
import string
import urllib

#######################################

def convert(obj, cls):
    '''
    Attempts to convert the given object to the given class. This is mostly
    useful for conversions between TextBased types, e.g.

    >>> convert('hello there', Url)
    Url('hello%20there')

    Rules for conversion are as follows:
     1. If the object is already an instance of the given class, no conversion
         is performed.
     2. If the object defines obj._convert_to_(cls), that is called and the
         result returned.
     3. If rule 2 raised NotImplementedError or did not apply, then if the
         class defines cls._convert_from_(obj), that is called and the result
         returned.
    '''
    if not inspect.isclass(cls):
        raise TypeError('%r is not a class' % (cls,))
    if isinstance(obj, cls):
        return obj
    try:
        if hasattr(obj, '_convert_to_'):
            return obj._convert_to_(cls)
    except NotImplementedError:
        pass
    try:
        if hasattr(cls, '_convert_from_'):
            return cls._convert_from_(obj)
    except NotImplementedError:
        pass
    raise TypeError('%r cannot be automatically converted to %s' % (obj,
            cls.__name__))

def raw(obj):
    '''
    Returns the raw string or unicode object underlying the given object, or
    raises TypeError if a raw string does not make sense.
    The raw string is obtained by calling obj._raw_() if such is defined.
    '''
    if hasattr(obj, '_raw_'):
        return obj._raw_()
    if isinstance(obj, (str, unicode)):
        return obj
    raise TypeError('%r has no raw striing' % (obj,))

#######################################

class Formatter(string.Formatter):
    '''
    An extension of the string.Formatter class which has the following
    features:
     * Any object can be used as a format template (not just strings) provided
        that the object defines a _parse_format_(f) method that accepts the
        formatter as its only argument and returns an iterable that contains
        tuples of the form:
            (literal_text, field_name, format_spec, conversion)
        OR
            (literal_text, field_name, format_spec, (conversion, field_conv))
        field_conv is a type and, if specified, the object to be formatted
        will be converted to that type using convert(). 
     * Objects that act as format templates may define a method
       _finish_format_(s) which accepts the raw string s resulting from the
       format, and manipulates it in some way before returning the new format
       result.
     * Field types may be registered using formatter.registertype(). Once a
        type is registered with the formatter, it may be used as part of the
        field specifiers.
     * Strings that are used as format strings can specify field types.
        For instance, {0:/Plain} indicates that the field is expecting plain
        text, and {0:20/Url} indicates that the field is expecting a url which
        will then be formatted to have a width of at least 20 characters. This
        requires that a type has been registered with the key 'Url'.
    '''
    def __init__(self):
        self.types = {'Plain': PlainText}

    def registertype(self, type, key=None):
        '''
        Registers the given field type with the given key for use in field
        specifiers. For instance, if you register a type with the key 'Url',
        you can convert to that type using {0:/Url}. If no key is specified,
        type.key is used.
        Returns the type itself so that this method may be used as a decorator.
        '''
        if key is None:
            key = type.key
        if key in self.types:
            raise KeyError('key %r already in use' % (key,))
        self.types[key] = type
        return type

    def parse(self, format_string):
        '''
        Returns an iterable that contains tuples of the form:
            (literal_text, field_name, format_spec, conversion)
        OR
            (literal_text, field_name, format_spec, (conversion, field_conv))
        See the docs for string.Formatter.parse for the definitions of
        literal_text, field_name, format_spec, conversion.

        If field_conv is given, it must be a type, which will be converted to
        using convert().
        '''
        if hasattr(format_string, '_parse_format_'):
            return format_string._parse_format_(self)
        return self.parse_string(format_string)

    def parse_string(self, format_string):
        '''
        The same as parse(), but requires that the format_string is actually a
        string. 
        '''
        if not isinstance(format_string, (str, unicode)):
            raise TypeError('%r is not a valid format string' %
                    (format_string,))
        for literal_text, field_name, format_spec, conversion in (
                format_string._formatter_parser()):
            if format_spec:
                format_spec, sep, field_conversion = format_spec.rpartition(
                        '/')
                if sep:
                    if field_conversion not in self.types:
                        raise KeyError('Unknown field type: %r' %
                                (field_conversion,))
                    yield (literal_text, field_name, format_spec, (conversion,
                            self.types[field_conversion]))
                    continue
            yield literal_text, field_name, format_spec, conversion

    def _vformat(self, format_string, args, kwargs, used_args, recursion_depth):
        if recursion_depth < 0:
            raise ValueError('Max string recursion exceeded')
        result = []
        for literal_text, field_name, format_spec, conversion in (
                self.parse(format_string)):
            # output the literal text
            if literal_text:
                result.append(literal_text)

            # if there's a field, output it
            if field_name is not None:
                result.append(self.render_field(field_name, format_spec,
                        conversion, args, kwargs, used_args, recursion_depth))

        result = ''.join(result)
        if hasattr(format_string, '_finish_format_'):
            result = format_string._finish_format_(result)
        return result

    def render_field(self, field_name, format_spec, conversion, args, kwargs,
            used_args, recursion_depth):
        '''
        Formats the given field and returns the raw formatted string.
        '''
        if isinstance(conversion, tuple):
            conversion, field_conversion = conversion
        else:
            field_conversion = None

        obj, arg_used = self.get_field(field_name, args, kwargs)
        used_args.add(arg_used)

        # do any conversion on the resulting object
        obj = self.convert_field(obj, conversion, field_conversion)

        # expand the format spec, if needed
        format_spec = self._vformat(format_spec, args, kwargs, used_args,
                recursion_depth - 1)

        # format the object and append to the result
        return self.format_field(obj, format_spec)

    def convert_field(self, value, conversion, field_conversion):
        '''
        Perform the conversion on the given value. If conversion is specified,
        it is None or a string ('r' or 's') as for
        string.Formatter.convert_field. If field_conversion is not None,
        it must be a type and the value is converted to that type using
        convert().
        '''
        if conversion == 'w':
            value = raw(value)
        else:
            value = string.Formatter.convert_field(self, value, conversion)
        if field_conversion is None:
            return value
        return raw(convert(value, field_conversion))

class Formattable(object):
    '''
    Base class for objects which can be used as format templates.
    '''
    __slots__ = ()  # In case subclasses need this.

    def format(self, *args, **kwargs):
        '''
        Uses this object as a format template with the given substitution
        variables. The default formatter will always be used by this method.
        If you want to use another formatter, use formatter.format(template,
        *args, **kwargs).
        Equivalent to formatter.format(v, *args, **kwargs).
        '''
        return formatter.vformat(self, args, kwargs)

    def _parse_format_(self, formatter):
        raise NotImplementedError()
        
#######################################

class TextBased(Formattable):
    '''
    Base class for objects which are wrappers around some raw text.
    '''
    __slots__ = ('raw',)  # In case subclasses need this.

    def __init__(self, raw):
        if not isinstance(raw, (str, unicode)):
            raise TypeError('raw must be str or unicode')
        if not self.valid_raw_text(raw):
            raise ValueError('%r is not valid raw text for %s' % (raw,
                    self.__class__.__name__))
        self.raw = raw

    def __str__(self):
        return raw(convert(self, PlainText))

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.raw)

    def _raw_(self):
        return self.raw

    def _parse_format_(self, formatter):
        for literal_text, field_name, format_spec, conversion in (
                formatter.parse_string(self.raw)):
            if field_name and not isinstance(conversion, tuple):
                conversion = (conversion, self.__class__)
            yield literal_text, field_name, format_spec, conversion

    def _finish_format_(self, result):
        return self.__class__(result)

    @staticmethod
    def valid_raw_text(text):
        '''
        Returns True if the given text is valid raw text for this class.
        '''
        return False

class PlainText(TextBased):
    '''
    Represents text that is designed to be user-readable.
    '''
    __slots__ = ()

    @classmethod
    def _convert_from_(cls, other):
        if isinstance(other, (str, unicode)):
            return PlainText(other)
        raise NotImplementedError

    def _convert_to_(self, cls):
        if cls in (str, unicode):
            return cls(self.raw)
        raise NotImplementedError

    @staticmethod
    def valid_raw_text(text):
        '''
        Any valid ASCII or unicode is valid as raw text.
        '''
        if isinstance(text, unicode):
            return True
        if isinstance(text, str):
            try:
                text.decode('ascii')
            except UnicodeDecodeError:
                return False
            return True
        return False

#######################################

# Define the default formatter and default formatting functions.
formatter = Formatter()
format = formatter.format
vformat = formatter.vformat

#######################################
# What follows are simple examples.
#######################################

@formatter.registertype
class Url(TextBased):
    '''
    Simple example type that performs url quoting on plain text.

    >>> convert('path/bits', Url)
    Url('path%2Fbits')
    >>> convert('path bits', Url)
    Url('path%20bits')
    '''
    key = 'Url'
    @classmethod
    def _convert_from_(cls, other):
        if isinstance(other, PlainText):
            other = raw(other)
        if isinstance(other, (str, unicode)):
            return Url(urllib.quote(other, ''))
        raise NotImplementedError

    @staticmethod
    def valid_raw_text(text):
        return True

@formatter.registertype
class UrlPath(TextBased):
    '''
    Simple example type that performs url quoting on plain text, but does not
    quote slashes.

    >>> convert('path/bits', UrlPath)
    UrlPath('path/bits')
    >>> convert('path bits', UrlPath)
    UrlPath('path%20bits')

    This class is directly convertible to and from the Url class.
    >>> convert(Url('path'), UrlPath)
    UrlPath('path')
    >>> convert(UrlPath('path'), Url)
    Url('path')
    '''
    key = 'UrlPath'

    @classmethod
    def _convert_from_(cls, other):
        if isinstance(other, Url):
            return cls(raw(other))

        if isinstance(other, PlainText):
            other = raw(other)
        if isinstance(other, (str, unicode)):
            return UrlPath(urllib.quote(other))
        raise NotImplementedError

    def _convert_to_(self, cls):
        if cls == Url:
            return cls(self.raw)
        raise NotImplementedError

    @staticmethod
    def valid_raw_text(text):
        return True

#######################################

if __name__ == '__main__':
    t = PlainText('There shall be {0}, but not until {foo}.')
    print t.format('food', foo='after lunch')
    print t.format('laughter', foo='it stops raining')
    print

    t1 = Url('http://{base:/UrlPath}/path?{key1}=val1;key2={val2}')
    print repr(t1.format(base='localhost', key1='a key', val2='a val'))
    print repr(t1.format(base='localhost/pi', key1='key1', val2='val2'))
    print repr(t1.format(base=convert('localhost/pi', Url), key1='key1',
            val2='val2'))
    print repr(t1.format(base=UrlPath('localhost:8080'), key1='key1', val2='val2'))
    print

    t2 = PlainText('The url is {0:/Url}, but do not visit it.')
    print t2.format('http://something')
    print t2.format(Url('http://something'))
